from __future__ import division
from lib.dynamic_objects import DynamicObjectStore, OBJECT_TYPES
from lib.snowball import FlyingSnowball

from lib.brain import BehaviorTree, enemies_in_range

from lib.unit import Unit

import random
import rabbyt
from pyglet.text import Label
import snowui


class Igloo(DynamicObjectStore):
    """
    Stuff the Building class does:

        * stores objects (by subclassing from ObjectStore)

        * can be owned by a player (takeover logic is here...)

        * Penguins stay inside.
    """
    accessable_by_sbs = ["player", "map_xy"]
    name = "igloo"
    _can_takeover = False
    size = (4,4)
    def __init__(self, scene, map_xy, player=None):
        DynamicObjectStore.__init__(self, scene)
        self.scene = scene
        assert(type(map_xy) == tuple)
        self.map_xy = map_xy
        self.takeover_player = None
        self.takeover_finished_time = 0
        self.units_inside = set()
        self.tiles = []

        self._construction = 1

        # TODO abstract to work for different types of buildings.

        self.sprite = rabbyt.Sprite("data/"+self.name+".png")
        self.sprite.shape.bottom = 0
        self.sprite.xy = scene.map.display.to_pixel(self.map_xy)
        self.sprite.x += self.size[0]/2*32
        self.y = self.sprite.y + self.size[1] * 32

        self.player_sprite = rabbyt.Sprite("data/marker.png", xy=self.sprite.xy)
        self.player_sprite.right = self.sprite.x+self.size[0]/2*32
        self.player_sprite.bottom = self.sprite.bottom

        self._player = None
        self.player = player

        self._ctrl_attached_mission = None
        self.glow = rabbyt.Sprite("data/"+self.name+"_glow.png")
        self.glow.xy = self.sprite.xy
        self.glow.shape.bottom = self.sprite.shape.bottom
        self.glow.scale_x = self.sprite.attrgetter("scale_x")
        self.glow.scale_y = self.sprite.attrgetter("scale_y")

        self.label = None

        self._gui_pane = None

    @property
    def can_takeover(self):
        if self._construction >= 1:
            return self._can_takeover
        return False

    @property
    def gui_pane(self):
        if self._gui_pane:
            # Creating this gui pane is a little expensive (especially when you
            # do many at once). So they are done on demand to save loading time.
            return self._gui_pane

        self._gui_pane = snowui.Widget(texture="data/building_info.png", rx=20,
            ry=0, bounds="rect")
        self._gui_pane.add(snowui.Text("Igloo", rx=20, ry=250, font_size=25))

        self._gui_type_store = {}
        for i, (name, cls) in enumerate(OBJECT_TYPES.items()):
            if name == "snowball": continue
            t = snowui.Text(name+": "+str(self.obj_counter[cls]), rx=20,
                ry=200 - 30*i, font_size=16)
            self._gui_type_store[cls] = t
            self._gui_pane.add(t)

        self._gui_pane.add(snowui.Text("New Penguin: (10 fish)", rx=20, ry=95, font_size=18))
        b = snowui.Button(icon=snowui.Text("New Worker", font_size=14,
            rx=50, ry=10), texture="data/button.png", rx=35, ry=10,
            bounds="rect", callback=self.create_penguin, cb_data="worker")
        self._gui_pane.add(b)

        b = snowui.Button(icon=snowui.Text("New Snowballer", font_size=14,
            rx=50, ry=10), texture="data/button.png", rx=35, ry=50,
            bounds="rect", callback=self.create_penguin, cb_data="snowballer")
        self._gui_pane.add(b)
        return self._gui_pane

    def create_penguin(self, type):
        if self.obj_counter[OBJECT_TYPES["fish"]] >= 10:
            for i in range(10):
                self.take_object(OBJECT_TYPES["fish"])
            u = Unit(self.scene, type, self.map_xy, self.player)
            self.scene.units.append(u)
            u.enter_building(self)
            #TODO have goto flag
            u.set_job((self.map_xy[0], self.map_xy[1]-2), 1, force_walk=True)

            s = rabbyt.Sprite("data/notify_new_penguin.png", x=self.sprite.x-30)
            s.alpha = rabbyt.ease_in(end=0, dt=2)
            s.y = rabbyt.ease_out(start=self.sprite.y+80,
                end=self.sprite.y+80+70, dt=2)
            self.scene.ui_imprents.add(s)

    def amount_changed(self, cls):
        if hasattr(self, "_gui_type_store"):
            self._gui_type_store[cls].text = cls.name+": "+str(self.obj_counter[cls])

    def attach_label(self, string):
        if not self.label:
            center_x = (self.sprite.left+self.sprite.right)/2
            # FIXME? under certain circumstances using TSCu_Comic will cause a
            # segfault.
            self.label = Label(string, font_name="SerafettinCartoon",
                    x=center_x, y=self.sprite.top, anchor_x="center",
                    color=(0,0,0,255))
        else:
            self.label.text = string


    def attach_mission(self, mapname):
        self._ctrl_attached_mission = mapname

    def add_to_display(self):
        self.node = self.scene.sprite_sets.get(self.sprite.x, self.sprite.y)
        self.node.add(self)

    def set_player(self, player):
        if player:
            self.player_sprite.rgb = player.color
        self._player = player
    player = property(lambda self:self._player, set_player)

    def set_map_x(self, x):
        self.map_xy[0] = x
    def set_map_y(self, y):
        self.map_xy[1] = y

    map_x = property(lambda self:self.map_xy[0], set_map_x)
    map_y = property(lambda self:self.map_xy[1], set_map_y)

    def set_construction(self, a):
        self._construction = a
        if a < .4:
            self.sprite.scale = rabbyt.ease_out(end=.4, dt=1, method="bounce")
            self.sprite.alpha = rabbyt.lerp(start=.8, end=.4, dt=1, extend="reverse")
        else:
            self.sprite.scale = rabbyt.ease_out(end=a, dt=1, method="bounce")
        if a >= 1:
            self.sprite.alpha = 1
    construction = property(lambda self:self._construction, set_construction)

    def render(self):
        if self._ctrl_attached_mission or self.scene.ctrl.selected_building == self:
            self.glow.render()
        self.sprite.render()
        self.player_sprite.render()
        if self.label:
            self.label.draw()

    def takeover_started_feedback(self, takeover_player, time_left):
        # This function is called on the client to display feedback for a
        # takeover.
        self.takeover_player = takeover_player
        self.takeover_finished_time = self.scene.time + time_left
        print "Takeover started by", takeover_player
        if self.player:
            self.player_sprite.rgb = rabbyt.lerp(
                    self.player.color, takeover_player.color,
                    dt=1, extend="reverse")
            self.player_sprite.alpha = 1
        else:
            self.player_sprite.rgb = takeover_player.color
            self.player_sprite.alpha = rabbyt.lerp(0,1, dt=1, extend="reverse")

    def takeover_successful_feedback(self):
        self.player = self.takeover_player
        self.takeover_player = None
        self.takeover_finished_time = 0
        print "Takeover successful!"
        self.player_sprite.rgb = self.player.color
        self.player_sprite.alpha = 1

    def takeover_failed_feedback(self):
        self.takeover_player = None
        self.takeover_finished_time = 0
        print "Takeover failed!"
        if player:
            self.player_sprite.alpha = 1
            self.player_sprite.rgb = self.player.color
        else:
            self.player_sprite.alpha = 0

    def add_unit(self, unit):
        self.units_inside.add(unit)
        if unit.player not in (self.player, self.takeover_player):
            self.begin_takeover(unit.player)

    def remove_unit(self, unit):
        # This should already be called on both the client and the server by
        # the unit, so we don't need to send a message about it.
        self.units_inside.remove(unit)

    def begin_takeover(self, player):
        c = self.scene.clock
        if self.takeover_player:
            # If we were already in the middle of a takeover, we want to
            # cancel the previous one.
            c.unschedule(self.end_takeover)
        self.takeover_player = player
        takeover_time = 2.0 # FIXME load the takeover time from an ini?
        self.takeover_finished_time = self.scene.time + takeover_time
        c.schedule_once(self.end_takeover, takeover_time)

    def end_takeover(self, dt):
        # This method is called by the scheduler after the takeover time has
        # completed.  (In other words, this handles a successful takeover.)

        for k,v in self.obj_counter.items():
            if self.player:
                self.player.points -= k.points*v
            self.takeover_player.points += k.points*v

        self.takeover_finished_time = 0
        self.player = self.takeover_player
        self.takeover_player = None

        # FIXME what do we do about any of the other player's units in here?



class DefendJob(object):
    def __init__(self, tower):
        self.tower = tower

    def do(self):
        if self.tower.construction >= 1:
            enemies = enemies_in_range(self.tower)
            for i in range(self.tower.get_setting("num_shots_per_round")):
                if enemies:
                    self.tower.throw_at(random.choice(enemies))

        return self.tower.get_setting("reload_time")



class Tower(Igloo):
    name = "tower"
    size = (3,2)
    _can_takeover = True

    settings = {"throw_range":10, "num_shots_per_round":1, "reload_time":1}

    accessable_by_sbs = Igloo.accessable_by_sbs + ["reload_time", "num_shots_per_round"]

    def __init__(self, scene, map_xy, player=None):
        Igloo.__init__(self, scene, map_xy, player=player)
        self.job = None

        self.bt = BehaviorTree(self.behavior_root)
        self.scene.clock.schedule_once(self.process, random.random())

        # Compatibility with unit AI
        self.mood = None

    @property
    def gui_pane(self):
        if self._gui_pane:
            # Creating this gui pane is a little expensive (especially when you
            # do many at once). So they are done on demand to save loading time.
            return self._gui_pane

        self._gui_pane = snowui.Widget(texture="data/building_info.png", rx=20,
            ry=0, bounds="rect")
        self._gui_pane.add(snowui.Text("Tower", rx=20, ry=250, font_size=25))

        self._gui_type_store = {}
        for i, (name, cls) in enumerate(OBJECT_TYPES.items()):
            if name == "snowball": continue
            t = snowui.Text(name+": "+str(self.obj_counter[cls]), rx=20,
                ry=200 - 30*i, font_size=16)
            self._gui_type_store[cls] = t
            self._gui_pane.add(t)

        return self._gui_pane

    def get_setting(self, option):
        return self.settings[option]

    def process(self, dt):
        if self.bt:
            res = self.bt.execute()
            if res == "fail":
                print "Warning! root behavior tree failed!"
            elif res == "success":
                print "Warning! root behavior tree exited!"
            else:
                self.scene.clock.schedule_once(self.process, res)
                return
        self.scene.clock.schedule_once(self.process, .2)

    def behavior_root(self):
        if self.job:
            return self.job.do()
        else:
            self.job = DefendJob(self)
            return .3

    def throw_at(self, target, dt=None):
#         if not dt:
#             print "warning: tower.throw_at dt is none"

        if hasattr(target, 'map_xy'):
            map_xy = target.map_xy
        else:
            map_xy = target

        snb = FlyingSnowball(self, map_xy)

class Tower2(Tower):
    name = "tower2"
    size = (3,2)
    settings = {"throw_range":10, "num_shots_per_round":6, "reload_time":4}

BUILDINGS = {"igloo":Igloo, "tower": Tower, "tower2":Tower2}
