from __future__ import division
import sys
import rabbyt
from pyglet import image
from pyglet.gl import *
from pyglet.clock import ClockDisplay

from lib.map import MapWrapper
from lib.scene import Scene
from lib.unit import Unit
from lib.player import Player
from lib.pixnodeset import PixNodeSet
from lib.ctrl import GameControl
from lib.playerai import PlayerAi
from lib import snowflakes

from lib import profiler as p

class GameScene(Scene):

    textures = {
        "snow": image.load("data/snow.png"),
        #"ice": image.load("data/ice.png"),
        "snowflakes": image.ImageGrid(image.load("data/snowparticles.png"),
                        rows=2, columns=2),
        "bear": image.ImageGrid(
                image.load("data/bear.png"), rows=8,
                columns=8, item_width=64, item_height=64),
    }
    for n in "walk gathersnow throw gatherfish unloadfish walkfish unloadcrystal walkcrystal workergatherfish workerunloadcrystal workerunloadfish workerwalk workerwalkfish snowballergathersnow snowballerthrow snowballerwalk".split():
        textures[n+"_anim"] = image.load("data/"+n+".png")
    textures["workerwalkcrystal_anim"] = textures["workerwalkfish_anim"]


    def __init__(self, scene_handler, mapname):
        assert p.push("Scene")
        self.scene_handler = scene_handler
        self.ctrl = GameControl(self)
        Scene.__init__(self)
        assert p.pop()
        self.setup(mapname)

        if "debug" in sys.argv:
            self.fps_display = ClockDisplay(clock=self.clock)

    def setup(self, mapname):
        self.dynamic_objects = []
        self.units = []
        self.map = None
        self.player = None
        self.players = []
        self.buildings = []
        self.flying_snowballs = []
        self.snowball_imprents = set()
        self.ui_imprents = set()

        self.paths = []
        self.script_areas = {}

        assert p.push("Misc")
        self.sprites = {
            "selected": rabbyt.Sprite("data/selected.png"),
        }
        self.sprite_sets = PixNodeSet(size=256)
        self.sprite_sets_staticmo = PixNodeSet(size=256)

        self.darkness_overlay = rabbyt.Sprite(shape=(self.view.width,
                self.view.height), rgba=(0,0,0,0))
        self.darkness_overlay.bottom = 0
        self.darkness_overlay.left = 0

        self.flash_panel = rabbyt.Sprite(shape=(self.view.width,
                self.view.height), rgba=(1,1,1,0))
        self.flash_panel.bottom = 0
        self.flash_panel.left = 0

        self.snowflakes = snowflakes.Snowflakes(self)
        assert p.pop()

        assert p.push("Ctrl")
        self.ctrl.setup()
        assert p.pop()

        assert p.push("Map")
        self.load_map(mapname)
        assert p.pop()

        assert p.dump("Game Loading")


    def load_map(self, mapname):
        self.map = MapWrapper(self, mapname)
        self.map.load()

        self.day_cycle_time = self.map.model.day_cycle_time
        if self.map.use_day_night:
            self.clock.schedule_interval(self.map.model.day_cycle, self.day_cycle_time/2)
        self.percent_day_done = .5
        self.map.model.day_cycle(0)

        colors = [(1,0,0), (0,1,0), (0,0,1), (1,1,0), (0,1,1), (1,0,1),
                  (.2,.7,0), (.7,.2,0), (.2, 0, .7)]
        for i in range(self.map.model.number_players):
            self.add_player(colors[i], "AI"+str(i))
        if self.players:
            self.player = self.players[0]
            self.player.name = "Me"
        self.view.xy = self.map.display.to_pixel(self.buildings[0].map_xy)

        for o in self.map.display.static_map_objects_renderable:
            node = self.sprite_sets_staticmo.get(*o.xy)
            node.add(o)


        assert p.push("Offscreen map")
        self.map.display.load_offscreen_map()
        assert p.pop()

        if self.map.config["map"]["ai"] == "Default":
            for player in self.players[1:]:
                ai = PlayerAi(self, player)
                self.clock.schedule_interval(ai.run, 5)

        self.map.model.load_script()

    def add_player(self, color, name):
        player = Player(name=name, color=color)
        self.players.append(player)

        igloo = None
        for igloo in self.buildings:
            if not igloo.player:
                break
            igloo = None
        if not igloo:
            raise ValueError("No avalible starting igloo found for player")
        igloo.player = player
        player.igloo = igloo
        # Add units.
        for i in range(10):
            u = Unit(self, "snowballer", igloo.map_xy, player)
            self.units.append(u)
        for i in range(10):
            u = Unit(self, "worker", igloo.map_xy, player)
            self.units.append(u)

    def loop(self):
        Scene.loop(self)
        rabbyt.clear((0,0,0,0))
        assert p.push("Control")
        self.ctrl.update()
        assert p.pop()

        assert p.push("Render")
        assert p.push("Map")
        if self.view.zoom <= 2:
            self.map.render()
        else:
            # We are zoomed out to the point that we don't want to draw terrain
            # detail.
            self.map.render_paper()
        assert p.pop()

        assert p.push("Sprites")

        assert p.push("collect")
        to_render = []
        sets = self.sprite_sets.range(
                self.view.x-self.view.scale_w/2-32,
                self.view.y-self.view.scale_h/2-180,
                self.view.x+self.view.scale_w/2+32,
                self.view.y+self.view.scale_h/2)
        for s in sets:
            to_render.extend(s)
        assert p.pop()


        if self.view.zoom <= 2:
            assert p.push("collect static")
            sets = self.sprite_sets_staticmo.range(
                    self.view.x-self.view.scale_w/2-32,
                    self.view.y-self.view.scale_h/2-180,
                    self.view.x+self.view.scale_w/2+32,
                    self.view.y+self.view.scale_h/2)
            for s in sets:
                to_render.extend(s)
            assert p.pop()
            assert p.push("sort")
            to_render.extend(self.flying_snowballs)
            to_render.sort(key=lambda s: -s.y)
            assert p.pop()


        assert p.push("draw")
        if self.view.zoom <= 2:
            rabbyt.render_unsorted(self.snowball_imprents)
        rabbyt.render_unsorted(to_render)
        assert p.pop()
        assert p.pop() # Sprites

        time_till_cycle_end = (self.time+self.day_cycle_time/2) % self.day_cycle_time
        time_till_cycle_end = self.day_cycle_time - time_till_cycle_end
        self.percent_day_done = time_till_cycle_end / self.day_cycle_time

        assert p.push("Snowflakes")
        self.ctrl.gui_view.set_viewport()
        # Draw snow falling down.
        #if self.percent_day_done > .5:
        self.snowflakes.run(True)
        #else:
            #self.snowflakes.run(False)
        assert p.pop()

        assert p.push("Darkness Overlay")
        if self.map.use_day_night:
            # Draw darkness overlay
            darkness = self.percent_day_done
            # Flip lighting halfway through so we get a complete day/night cycle
            # like we want.
            if darkness < .5:
                darkness = 1 - darkness

            # We want to make it so it's lighter longer than it is darker and so it
            # doesn't get too dark.
            darkness -= .8
            darkness = max(0, min(darkness, .3))

            self.darkness_overlay.alpha = darkness
            self.darkness_overlay.render()

        if self.flash_panel.alpha > 0:
            glBlendFunc(GL_SRC_ALPHA, GL_ONE)
            self.flash_panel.render()
            glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

        assert p.pop()

        assert p.push("UI")
        self.view.set_viewport()
        assert p.push("Imprints")
        rabbyt.render_unsorted(self.ui_imprents)
        assert p.pop() # Imprints
        assert p.pop() # UI
        assert p.pop() # Render
        assert p.push("Render")
        assert p.push("UI")
        self.ctrl.draw_select_box()
        self.ctrl.draw_gui()
        assert p.pop()
        assert p.push("flip")
        if "debug" in sys.argv:
            self.fps_display.draw()
        self.scene_handler.window.flip()
        assert p.pop()
        assert p.pop() # Render

        assert p.push("Cleanup")
        to_r = set()
        for s in self.snowball_imprents:
            if s.alpha == 0:
                to_r.add(s)
        self.snowball_imprents.difference(to_r)

        to_r = set()
        for s in self.ui_imprents:
            if s.alpha == 0:
                to_r.add(s)
        self.ui_imprents.difference(to_r)
        assert p.pop()

        return self._return