from __future__ import division

import Image
import os, sys
from pyglet.gl import *
from pyglet import image
import ctypes
import rabbyt

from lib import terrain
from lib.path import Path, Stream
from lib import dynamic_objects
from lib import sbscript

from lib import building


class MapDisplay(object):
    """
    Strictly for displaying a MapModel instance.
    """
    tile_size = 32
    elevation_multiplier = 10

    def __init__(self, scene, model):
        self.scene = scene
        self.model = model

        #w = model.size[0]*32
        #h = model.size[1]*32
        #self.old_paper = rabbyt.Sprite("data/old_paper.jpg", shape=(0,w,h,0))

        self.offscreen_chunks = {}

        self.static_map_objects_renderable = []
        for smo in self.model.static_map_objects:
            s = StaticMapObjectView(scene, smo)
            self.static_map_objects_renderable.append(s)

        f = open(os.path.join("data", "maps", self.model.name, "paths.txt"), "r")
        for l in f.readlines():
            if ":" not in l:
                continue
            typ, l = l.split(":")
            points = []
            for xy in l.split(" "):
                if xy and xy != "\n":
                    points.append(tuple([int(i) for i in xy.split(",")]))
            # hack...
            if typ == "stream":
                scene.paths.append(Stream(points))
            else:
                scene.paths.append(Path(points, typ))
        f.close()

    def load_terrain(self):
        self.snow_terrain = terrain.Terrain(None, "all", self)
        #self.ice_terrain = terrain.Terrain(None, "ice", self)

    def render(self):
        glEnable(GL_TEXTURE_2D)
        v = self.scene.view
        ts = self.tile_size
        vw, vh = v.scale_w, v.scale_h
        tx,ty = (((v.x-vw/2)//ts, v.x//ts+vw/2//ts+2),
                ((v.y-vh/2)//ts-8, v.y//ts+vh/2//ts+2))

        glBindTexture(GL_TEXTURE_2D, self.scene.textures["snow"].texture.id)
        self.snow_terrain.draw(tx,ty)
        #glBindTexture(GL_TEXTURE_2D, self.scene.textures["ice"].texture.id)
        #self.ice_terrain.draw(tx,ty)

        for p in self.scene.paths:
            p.render()

    def load_offscreen_map(self):
        # TODO: Add this to loading time
        map_w = self.model.size[0]*32
        map_h = self.model.size[1]*32

        chunk_res = 2048
        chunk_tex_res = 512
        num_chunks = (map_w//chunk_res+1, map_h//chunk_res+1)

        glClearColor(0,0,0,0)
        for x in range(num_chunks[0]):
            for y in range(num_chunks[1]):
                tex = image.Texture.create_for_size(GL_TEXTURE_2D,
                        chunk_tex_res, chunk_tex_res, GL_RGB)

                l = x*chunk_res
                t = y*chunk_res+chunk_res
                r = x*chunk_res+chunk_res
                b = y*chunk_res
                glMatrixMode(GL_PROJECTION)
                glLoadIdentity()
                glViewport(0,0,chunk_tex_res,chunk_tex_res)
                glOrtho(l, r, b, t, -1000, 1000)
                glMatrixMode(GL_MODELVIEW)
                glLoadIdentity()
                glClear(GL_COLOR_BUFFER_BIT)

                # Draw terrain
                glBindTexture(GL_TEXTURE_2D, self.scene.textures["snow"].texture.id)
                self.snow_terrain.draw(draw_all=True)
                # Draw paths
                for p in self.scene.paths:
                    p.render()
                # Draw static map objects
                to_render = []
                sets = self.scene.sprite_sets_staticmo.nodes.values()
                for s in sets:
                    to_render.extend(s)
                to_render.sort(key=lambda s: -s.y)
                rabbyt.render_unsorted(to_render)

                glBindTexture(GL_TEXTURE_2D, tex.id)
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE)
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE)
                glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, 0, 0,
                        chunk_tex_res, chunk_tex_res, 0)

                self.offscreen_chunks[(x,y)] = rabbyt.Sprite(tex,
                        x=x*chunk_res, y=y*chunk_res,
                        shape=(0,chunk_res, chunk_res, 0))

        glViewport(0,0,self.scene.view.width,self.scene.view.height)

    def render_paper(self):
        for s in self.offscreen_chunks.values():
            s.render()

    # TODO Move to map control
    def from_pixel(self, (x, y)):
        """
        returns the tile coordinates matched by the given pixel coordinates,
        taking elevation into account.

        The pixel coordinates are relative to the map, NOT the view.
        """
        s = self.tile_size
        tx,ty = (max(0, (x // s) * s // s), max(0, (y // s) * s // s))
        tx = min(max(0,tx), self.model.size[0]-1)
        ty = min(max(0,ty), self.model.size[1]-1)

        match = (tx,ty)
        for i in reversed(range(0,10)):
            try:
                t = (tx,max(0,ty-i))
                if y < self.to_pixel(t)[1]+self.tile_size/2:
                    return t
            except KeyError:
                pass
        return match

    def to_pixel(self, (x,y)):
        x = min(max(0,x), self.model.size[0]-1)
        y = min(max(0,y), self.model.size[1]-1)
        return x*self.tile_size+self.tile_size/2, y*self.tile_size+\
                self.tile_size/2 +\
                self.model.tiles[(x,y)].elevation*self.elevation_multiplier




class StaticMapObjectView(rabbyt.Sprite):
    def __init__(self, scene, model):
        self.scene = scene
        model.display = self
        self.model = model
        rabbyt.Sprite.__init__(self, "data/"+model.name+str(model.variation)+".png")
        self.right = model.tile.x*MapDisplay.tile_size+32
        self.bottom = model.tile.y*MapDisplay.tile_size +\
                model.tile.elevation*MapDisplay.elevation_multiplier