from __future__ import division
import random, os
from lib import profiler as p

import numpy

from pyglet.gl import *

import ctypes

class Terrain(object):
    def __init__(self, image, ttype, map_display, light_image=None):
        self.vsize = (map_display.model.size[0]+1, map_display.model.size[1]+1)
        # Texture size
        tsize = 256

        assert p.push("Load data")
        cache_path = os.path.join("data", "maps", map_display.model.name)
        if os.path.exists(os.path.join(cache_path, "cache.npz")):
            cache = numpy.load(os.path.join(cache_path, "cache.npz"))
            self.v = cache["v"]
            self.v_c = cache["v_c"]
            self.v_t = cache["v_t"]

            self.using_cache = True
        else:
            # Vertex positions:
            self.v = numpy.zeros(self.vsize+(2,), dtype=numpy.float32)
            # Vertex color:
            self.v_c = numpy.zeros(self.vsize+(4,), dtype=numpy.float32)
            # Vertex texture coords:
            self.v_t = numpy.zeros(self.vsize+(2,), dtype=numpy.float32)

            self.using_cache = False
        assert p.pop()

        # Generate arrays if the cache doesn't exist (or TODO needs updating)
        if not self.using_cache:
            for vy in xrange(self.vsize[1]):
                for vx in xrange(self.vsize[0]):
                    assert p.push("Query")
                    if (vx,vy) in map_display.model.tiles:
                        tile1 = map_display.model.tiles[(vx, vy)]
                        m = tile1.elevation*map_display.elevation_multiplier
                    else:
                        m = 0
                    assert p.pop()

                    assert p.push("Vertexs")
                    self.v[vx, vy] = vx*map_display.tile_size, vy*map_display.tile_size + m
                    self.v_t[vx, vy] = vx*map_display.tile_size/tsize, vy*map_display.tile_size/tsize
                    assert p.pop()

                    assert p.push("Lighting")
                    self.v_c[vx, vy] = 1,1,1,1
                    if light_image:
                        try:
                            r,g,b,a = light_image.get_at((vx,vy))
                            self.v_c[vx, vy, :3] = r/255, g/255, b/255
                        except IndexError:
                            pass
                    else:
                        # Elevation shadows
                        try:
                            tile1 = map_display.model.tiles[(vx, vy)]
                            tile2 = map_display.model.tiles[(vx-1, vy-1)]
                            d = (tile2.elevation - tile1.elevation) / 10
                            if d > 0:
                                d = 1-d
                            else:
                                d = 1.0
                        except KeyError:
                            d = 1.0

                        # Elevation highlights
                        try:
                            tile1 = map_display.model.tiles[(vx, vy)]
                            tile2 = map_display.model.tiles[(vx+1, vy+1)]
                            d2 = (tile2.elevation - tile1.elevation) / 10
                            d2 = max(d2, 0)
                        except KeyError:
                            d2 = 0.0
                        d += d2
                        d -= 0.1
                        self.v_c[vx, vy, :3] = d, d, d
                    assert p.pop()

        #if not self.using_cache:
        #    assert p.push("Save cache")
        #    # Save cache if necessary
        #    numpy.savez(os.path.join(cache_path, "cache.npz"), v=self.v,
        #            v_c=self.v_c, v_t=self.v_t)
        #    assert p.pop()

        #if ttype != "all":
            #assert p.push("Blending")
            #for vy in xrange(self.vsize[1]):
                #for vx in xrange(self.vsize[0]):
                    ## Find the average alpha from surrounding tiles
                    #total = 0
                    #num = 0
                    #for key in [(vx,vy), (vx-1, vy), (vx-1, vy-1), (vx, vy-1)]:
                        #try:
                            #tile = map_display.model.tiles[key]
                            #if tile.type == ttype:
                                #total += 1
                            #num += 1
                        #except KeyError: pass
                    #alpha = total/num
                    #if alpha == .5:
                        #alpha += random.random()*.8-.4
                    #self.v_c[vx, vy, 3] = alpha
            #assert p.pop()

        assert p.push("Step 2")
        self.indexes = numpy.zeros((map_display.model.size[0], map_display.model.size[1], 4),
                dtype=numpy.uint16)
        r = numpy.arange(self.vsize[0]*self.vsize[1]).reshape(self.vsize) \
                [:map_display.model.size[0], :map_display.model.size[1]]
        self.indexes[...,0] = r
        self.indexes[...,1] = r + self.vsize[1]
        self.indexes[...,2] = r + 1 + self.vsize[1]
        self.indexes[...,3] = r + 1
        assert p.pop()


    def draw(self, range_x=None, range_y=None, draw_all=False):
        if draw_all:
            indexes = self.indexes
        else:
            start_x = int(max(range_x[0], 0))
            end_x = int(min(range_x[1], self.vsize[1]))
            start_y = int(max(range_y[0], 0))
            end_y = int(min(range_y[1],self.vsize[1]))
            indexes = self.indexes[start_x:end_x,start_y:end_y,:].copy()

        glPushClientAttrib(GL_CLIENT_VERTEX_ARRAY_BIT)
        glEnableClientState(GL_VERTEX_ARRAY)
        glEnableClientState(GL_TEXTURE_COORD_ARRAY)
        glEnableClientState(GL_COLOR_ARRAY)
        glEnable(GL_TEXTURE_2D)
        # FIXME have the terrain know about it's texture!
        #glBindTexture(GL_TEXTURE_2D, self.image.texture.id)
        conv = numpy.ctypeslib.ndpointer().from_param
        glVertexPointer(2, GL_FLOAT, 0, conv(self.v))
        glTexCoordPointer(2, GL_FLOAT, 0, conv(self.v_t))
        glColorPointer(4, GL_FLOAT, 0, conv(self.v_c))
        glDrawElements(GL_QUADS, len(indexes.flat),
                GL_UNSIGNED_SHORT, conv(indexes))

        glPopClientAttrib()

