from __future__ import division
import random
import rabbyt
from lib.brain import *
from lib.pathfinding import Pathfinding
from lib.snowball import FlyingSnowball
from lib import events
from lib import dynamic_objects

from lib import profiler as p

class BaseUnit(object):
    accessable_by_sbs = ["last_hit_by", "player", "warmth", "max_warmth",
            "map_xy", "map_x", "map_y", "can_throw",  "cold_point", "speed"]
    cold_point = 5
    speed = .7

    def behavior_root(self):
        if self.warmth == 0:
            return 1
        if self.warmth <= self.cold_point:
            if self.holding and self.holding.name != "fish":
                # Go and unload current holding. Does automaticly.
                pass
            else:
                return find_fish_behavior(self)
        if self.job:
            return self.job.do()
        else:
            self.job = HoldPositionJob(self, self.map_xy, 5)
            return .3

    def __init__(self, scene, unit_type, pos, player=None):
        self.scene = scene

        self.unit_type = unit_type

        self.player = player
        self._map_xy = pos
        self.prev_map_xy = pos
        # (start of last move, end of last move)
        # Used to find the current progress between tiles.
        self.move_time_bounds = (0,0)

        self._max_warmth = 20
        self._warmth = self.max_warmth

        # Direction that the unit is moving in
        self.direction = (0,-1)

        self.can_throw = unit_type == "snowballer"
        self._is_gathering = False
        self.holding = None
        self.in_building = None

        self.bt = BehaviorTree(self.behavior_root)
        self.job = None
        self.scene.clock.schedule_once(self.process, random.random())

        self.scene.clock.schedule_interval(self.do_warm, 10)

        self.pathfinding = Pathfinding(self)

        # Last unit to hit self with a snowball.
        self.last_hit_by = None

        self.default_mood = {"throw_range":10, "gather_speed_modifier":1, "after_throw_wait":None,
                "throw_walk_throw":True}
        self.mood = self.default_mood

    def get_setting(self, option):
        if self.mood and option in self.mood:
            return self.mood[option]
        else:
            return self.default_mood[option]

    def enter_mood(self, mood, duration):
        self.mood = mood
        self.scene.clock.schedule_once(self.exit_mood, duration)
        ms = rabbyt.Sprite("data/expression_mad.png")
        ms.x = self.sprite.attrgetter('x') - 16
        ms.y = self.sprite.attrgetter('y') + 15
        if self.player:
            ms.rgb=self.player.color
        ms.alpha = rabbyt.ease_in(end=0, dt=duration)
        self.scene.ui_imprents.add(ms)

    def exit_mood(self, dt):
        self.mood = self.default_mood

    def chance_mood(self, chance):
        if self.warmth == 0:
            return
        if self.warmth < 10:
            chance += .5
        if random.random() <= chance:
            # TODO: Multiple mood levels
            m = {"throw_range":14, "gather_speed_modifier":.6, "after_throw_wait":.2,
                    "throw_walk_throw":False}
            self.enter_mood(m, 15)

    def do_warm(self, dt):
        # We make sure warmth is grater than 0 because we can't rely on
        # cold_point being grater than 0.
        if self.warmth < self.max_warmth and self.warmth > self.cold_point\
                and self.warmth > 0:
            self.warmth += 1

    def remove(self):
        self.scene.units.remove(self)
        self.scene.clock.unschedule(self.process)
        self.scene.clock.unschedule(self.do_warm)
        self.current_node.remove(self)
        self.pathfinding.current_node.remove(self)


    def _set_max_warmth(self, value):
        value = max(value, 1)
        self._max_warmth = value
        if self.warmth > value:
            self.warmth = value
    max_warmth = property((lambda self: self._max_warmth), _set_max_warmth)


    def _set_warmth(self, value):
        value = max(value, 0)
        if value == self._warmth:
            return
        self._warmth = value

        if value == 0:
            events.fire("frozen", self)
            if self.last_hit_by.player != self.player:
                self.last_hit_by.player.points += 1

        self.warmth_sprite.scale_y = rabbyt.lerp(end=self._warmth/self.max_warmth, dt=1)
        self.warmth_sprite.green = min(1, self._warmth/self.max_warmth*2)
        self.warmth_sprite.blue = min(1, (1-self._warmth/self.max_warmth)*2)
    warmth = property((lambda self: self._warmth), _set_warmth)

    def set_map_x(self, x):
        self.map_xy = (x,self.map_xy[1])
    def set_map_y(self, y):
        self.map_xy = (self.map_xy[0], y)

    map_x = property(lambda self:self.map_xy[0], set_map_x)
    map_y = property(lambda self:self.map_xy[1], set_map_y)

    def set_map_xy(self, (x, y)):
        if abs(x-self.map_xy[0]) > 1 or abs(y-self.map_xy[1]) > 1:
            # Was set by SBScript.
            self._map_xy = (x, y)

            self.pathfinding.update_node()
            self.sprite.xy = self.scene.map.to_pixel(self.map_xy)
            current_node = self.scene.sprite_sets.get(self.x, self.y)
            if not (self.current_node is current_node):
                self.current_node.remove(self)
                current_node.add(self)
                self.current_node = current_node
            self.sprite.alpha = rabbyt.lerp(start=0, end=1, dt=1)
        else:
            self._map_xy = (x, y)
        s = self.scene.map.tiles[(x,y)].script_area
        if s:
            s.last_player_over = self.player

    map_xy = property(lambda self:self._map_xy, set_map_xy)


    @property
    def walk_speed(self):
        return self.speed

    def is_moving(self):
        if self.scene.time < self.move_time_bounds[1]:
            return True
        else:
            return False

    #@networking.server_only
    def walk_to(self, map_xy):
        """
        Set the penguin walking to the next tile to get to map_xy.
        return the time when we will arrive.
        """
        # TODO check for walking off the map.
        self.move_time_bounds = (self.scene.time, self.scene.time+self.walk_speed)
        self.prev_map_xy = self.map_xy
        self.map_xy = map_xy
        self.pathfinding.walk_to(map_xy)
        return self.walk_speed

    #@networking.server_only
    def path_to(self, pos):
        return self.pathfinding.path_to(pos)

    def pick_up(self, obj):
        """
        Called by the AI.. assumes that it is legal to pick up object.
        """
        obj.set_held(self)

        self.holding = obj

        # Return how long it takes to pick it up.
        return 0.01#.5

    def unload(self, map_xy=None, object_store=None):
        if not map_xy and not object_store:
            map_xy = self.map_xy
        if self.holding:
            if object_store:
                # HACK should this really be here?
                #if hasattr(object_store, "construction"):
                #    # Check to see if we are constructing a building.
                #    if object_store.construction < 1:
                #        # TODO make this dependent on how much it takes to
                #        # construct the building
                #        object_store.construction += .2
                #        self.holding = None
                #        return .5

                if hasattr(object_store, "player") and\
                        object_store.player == self.player:
                    self.add_points(self.holding.points)
                self.holding.set_in_object_store(object_store)

            else:
                self.holding.set_on_ground(map_xy)
            self.holding = None
        dt = .5
        self.pathfinding.clear_history()
        return dt

    #@networking.server_only
    def eat(self):
        fish = self.holding
        dt = self.unload(self.map_xy)
        fish.destroy()
        self.warmth = self.max_warmth

        if self.unit_type == "snowballer" and isinstance(self.job, HoldPositionJob):
            if self.distance_to(self.job.target) > 30:
                # Old job position is to far away, cancel.
                self.job = None

        return dt


    def throw_at(self, target, dt=None):
        if hasattr(target, 'map_xy'):
            map_xy = target.map_xy
        else:
            map_xy = target
        snb = FlyingSnowball(self, map_xy)
        self.holding.destroy()
        self.holding = None
        return self.get_setting("after_throw_wait") or snb.dt



    #@networking.server_only
    def process(self, dt):
        if self.bt:
            res = self.bt.execute()
            if res == "fail":
                print "Warning! root behavior tree failed!"
            elif res == "success":
                print "Warning! root behavior tree exited!"
            else:
                self.scene.clock.schedule_once(self.process, res)
                return
        self.scene.clock.schedule_once(self.process, .2)

    #@networking.server_only
    def set_job(self, pos, stray_range, run=False, force_walk=False):
        # If force walk don't do gather etc
        t = self.scene.map.tiles[pos]
        self.pathfinding.clear_history()

        if self.unit_type == "worker" and not force_walk:
            for o in self.scene.dynamic_objects:
                if not o.held and t.xy == o.map_xy:
                    self.job = PickupObjectJob(self, o)
                    self.pathfinding.clear_history()
                    return "pickup"

            if t.static_map_object and t.static_map_object.gatherable:
                self.job = GatherJob(self,
                            dynamic_objects.OBJECT_TYPES[t.static_map_object.name],
                            starting_target=t.static_map_object)
                return "gather"
            elif t.building and t.building.player == self.player:
                #if t.building.construction < 1:
                #    self.job = BuildJob(self, t.building)
                #    return "build"
                #else:
                    self.job = EnterBuildingJob(self, t.building)
                    return "enter"
            else:
                self.job = HoldPositionJob(self, pos, stray_range)
                return "move"
        else:
            if t.building and (t.building.can_takeover or\
                    t.building.player == self.player):
                self.job = EnterBuildingJob(self, t.building)
                return "enter"
            else:
                self.job = HoldPositionJob(self, pos, stray_range, run)
                return "move"


    def distance_to(self, map_xy):
        # TODO make this more accurate.
        return abs(self.map_xy[0]-map_xy[0]) + abs(self.map_xy[1]-map_xy[1])


    def enter_building(self, building):
        # FIXME units should no longer be obstacles when inside a building.
        self.in_building = building
        building.add_unit(self)
        return .5

    def exit_building(self):
        self.in_building.remove_unit(self)
        self.in_building = None
        return .5


    def add_points(self, p):
        self.player.points += p

    def remove_points(self, p):
        self.player.points -= p



class Unit(BaseUnit):
    """ This is the graphical part of the unit. """

    SPRITE_DIRECTION_MAP = [(0,-1), (-1,-1), (-1,0), (-1, 1), (0,1), (1,1),
                            (1,0), (1,-1)]
    num_frames = 10

    def __init__(self, *args, **kwargs):
        super(Unit, self).__init__(*args, **kwargs)

        self.selected = False
        self.frame = 10
        self.last_frame = 10

        txy = self.scene.map.to_pixel(self.map_xy)
        self.sprite = rabbyt.Sprite(shape=(32,32),
                tex_shape=(0,32/256,32/512,0), xy=txy)
        self.hat_sprite = rabbyt.Sprite(shape=(32,32),
                tex_shape=(0,32/256,32/512,0), xy=txy)
        self.hat_sprite.x = self.sprite.attrgetter('x')
        self.hat_sprite.y = self.sprite.attrgetter('y')
        if self.player:
            self.hat_sprite.rgb=self.player.color

        self.icon = rabbyt.Sprite("data/penguin_icon.png")
        if self.player:
            self.icon.rgb = self.player.color
        self.icon.x = self.sprite.attrgetter('x')
        self.icon.y = self.sprite.attrgetter('y')

        self.current_node = self.scene.sprite_sets.get(self.x, self.y)
        self.current_node.add(self)

        self.warmth_sprite = rabbyt.Sprite(shape=(2,10), rgb=(0,1,0))
        self.warmth_sprite.shape.bottom = 0
        self.warmth_sprite.x = self.sprite.attrgetter('x') + 19
        self.warmth_sprite.y = self.sprite.attrgetter('y')
        self.warmth_sprite.scale_y = self.warmth/self.max_warmth

        self.animname = "walk_anim"


    def render(self):
        if self.scene.view.zoom > 4:
            self.icon.render()
            return

        if self.in_building:
            return

        if self.selected:
            s = self.scene.sprites["selected"]
            s.xy = self.sprite.xy
            s.render()

        if self.animname == "walk_anim" and not self.is_moving():
            self.frame = 10

        self.sprite.rgba = 1,1,1,1
        row = self.SPRITE_DIRECTION_MAP.index(self.direction)
        # We do 11 here because there is an extra frame on the end of each
        # row we have to account for.
        i = (7-row)*11+int(self.frame)

        self.sprite.texture = self.scene.textures[self.animname].texture.id
        self.sprite.u = int(self.frame)*32/512
        self.sprite.v = (7-row)*32/256
        try:
            self.hat_sprite.texture = self.scene.textures[self.unit_type+\
                    self.animname].texture.id
            self.hat_sprite.uv = self.sprite.uv
        except KeyError:
            pass

        if self.warmth != self.max_warmth and self.warmth_sprite.alpha == 0:
            self.warmth_sprite.alpha = rabbyt.lerp(end=1, dt=2)
        elif self.warmth == self.max_warmth and self.warmth_sprite.alpha == 1:
            self.warmth_sprite.alpha = rabbyt.lerp(end=0, dt=1)
        if self.warmth_sprite.alpha > 0:
            self.warmth_sprite.render()

        if self.warmth == 0:
            self.sprite.rgba = 0,0,1,.7

        self.sprite.render()
        self.hat_sprite.render()

    def _animate(self, dt):
        # Steps per tile walked.
#         sptw = 2
#         t = res/(sptw*10)

        # TODO: Time this properly
        self.frame += 1
        if self.frame > self.num_frames-1:
            self.frame = 0
            if self.animname in "throw_anim unloadfish_anim unloadcrystal_anim".split():
                self._done_animate()
    def _done_animate(self, *args):
        # FIXME: if we are going to walk again we don't want to reset the frame.
        self.last_frame = self.frame
        self.frame = 4
        if self.holding and self.holding.name == "fish":
            self.animname = "walkfish_anim"
        elif self.holding and self.holding.name == "crystal":
            self.animname = "walkcrystal_anim"
        else:
            self.animname = "walk_anim"
        self.scene.clock.unschedule(self._animate)

    def walk_to(self, map_xy):
        res = BaseUnit.walk_to(self, map_xy)
        txy = self.scene.map.to_pixel(map_xy)
        self.sprite.xy = rabbyt.lerp(end=txy, dt=self.walk_speed)

        self.frame = self.last_frame

        # We unschedule these to prevent any interfearance with frame inc.
        self.scene.clock.unschedule(self._animate)
        self.scene.clock.unschedule(self._done_animate)

        self.scene.clock.schedule(self._animate)
        self.scene.clock.schedule_once(self._done_animate, res)

        # Update self in map nodes.
        current_node = self.scene.sprite_sets.get(self.x, self.y)
        if not (self.current_node is current_node):
            self.current_node.remove(self)
            current_node.add(self)
            self.current_node = current_node

        return res

    @property
    def x(self):
        return self.sprite.x
    @property
    def y(self):
        return self.sprite.y

    def start_gathering(self, object_type, map_xy):
        if object_type.name == "snowball":
            self.animname = "gathersnow_anim"
        elif object_type.name == "fish":
            self.animname = "gatherfish_anim"
        elif object_type.name == "crystal":
            self.animname = "gatherfish_anim"
        self.frame = 0
        self.scene.clock.unschedule(self._animate)
        self.scene.clock.schedule(self._animate)
        self.scene.clock.schedule_once(self._done_animate, object_type.gather_time)

#     def pick_up(self, obj):
#         self.animname = "gathersnow_anim"
#         self.frame = 0
#         self.scene.clock.unschedule(self._animate)
#         self.scene.clock.schedule(self._animate)
#         self.scene.clock.schedule_once(self._done_animate, .7)
#         return BaseUnit.pick_up(self, obj)

    def throw_at(self, target, dt=None):
        self.animname = "throw_anim"
        self.frame = 0
        self.scene.clock.schedule(self._animate)

        return BaseUnit.throw_at(self, target, dt)


    def unload(self, map_xy=None, object_store=None):
        if self.holding.name == "fish":
            self.animname = "unloadfish_anim"
        if self.holding.name == "crystal":
            self.animname = "unloadcrystal_anim"
        self.frame = 0
        self.scene.clock.unschedule(self._animate)
        self.scene.clock.schedule(self._animate)
        return BaseUnit.unload(self, map_xy=map_xy, object_store=object_store)




class Bear(Unit):
    cold_point = 5
    speed = .4

    num_frames = 8

    def __init__(self, *args, **kwargs):
        BaseUnit.__init__(self, *args, **kwargs)

        self.selected = False
        self.frame = 0
        self.last_frame = 0

        txy = self.scene.map.to_pixel(self.map_xy)
        self.sprite = rabbyt.Sprite(shape=(64,64), xy=txy)

        self.current_node = self.scene.sprite_sets.get(self.x, self.y)
        self.current_node.add(self)

        self.warmth_sprite = rabbyt.Sprite(shape=(2,10), rgb=(0,1,0))
        self.warmth_sprite.shape.bottom = 0
        self.warmth_sprite.x = self.sprite.attrgetter('x') + 19
        self.warmth_sprite.y = self.sprite.attrgetter('y')
        self.warmth_sprite.scale_y = self.warmth/self.max_warmth

        self.animname = "bear"

    def render(self):
        if self.selected:
            s = self.scene.sprites["selected"]
            s.xy = self.sprite.xy
            s.render()

        if self.in_building:
            return

        if self.warmth == 0:
            self.sprite.rgba = 0,0,1,.7
        else:

            self.sprite.rgba = 1,1,1,1
            row = self.SPRITE_DIRECTION_MAP.index(self.direction)
            i = (7-row)*self.num_frames+int(self.frame)
            self.sprite.texture = self.scene.textures["bear"][i].texture

            if self.warmth != self.max_warmth and self.warmth_sprite.alpha == 0:
                self.warmth_sprite.alpha = rabbyt.lerp(end=1, dt=2)
            elif self.warmth == self.max_warmth and self.warmth_sprite.alpha == 1:
                self.warmth_sprite.alpha = rabbyt.lerp(end=0, dt=1)
            if self.warmth_sprite.alpha > 0:
                self.warmth_sprite.render()

        self.sprite.render()
